@@ -51,9 +51,25 @@ void leon_init(ambapp_core *core, int level)
 static int irqmp_initied = 0;
 static int eirqs[16];
 
+static void irqmp_add_eirq(struct irqmp_regs *cregs)
+{
+	int eirq;
+	int i;
+
+	eirq = (BYPASS_LOAD_PA(&cregs->mpstatus) >> 16) & 0xf;
+	for (i=0; eirq && i < 16; i++) {
+		if (!eirqs[i]) {
+			eirqs[i] = eirq;
+			break;
+		} else if (eirqs[i] == eirq) {
+			break;
+		}
+	}
+}
+
 int irqmp_dev(ambapp_dev *dev)
 {
-	int i, c, icsel, eirq;
+	int c, icsel, nctrl;
 	ambapp_apbdev *apbslv;
 	struct irqmp_regs *regs, *cregs;
 
@@ -66,21 +82,18 @@ int irqmp_dev(ambapp_dev *dev)
 	}
 
 	/* Record eirqs, so that grgpio can avoid collisions */
-	for (c = 0; c < 16; c++) {
-		/* Logic lifted from leon_kernel.c in linux kernel */
-		icsel = BYPASS_LOAD_PA(&regs->icsel[c/8]);
-		icsel = (icsel >> ((7 - (c & 0x7)) * 4)) & 0xf;
-		cregs = regs + icsel;
+	nctrl = BYPASS_LOAD_PA(&regs->ampctrl) >> 28;
+	if (nctrl) {
+		for (c = 0; c < 16; c++) {
+			/* Logic lifted from leon_kernel.c in linux kernel */
+			icsel = BYPASS_LOAD_PA(&regs->icsel[c/8]);
+			icsel = (icsel >> ((7 - (c & 0x7)) * 4)) & 0xf;
+			cregs = regs + icsel;
 
-		eirq = (BYPASS_LOAD_PA(&cregs->mpstatus) >> 16) & 0xf;
-		for (i=0; eirq && i < 32; i++) {
-			if (!eirqs[i]) {
-				eirqs[i] = eirq;
-				break;
-			} else if (eirqs[i] == eirq) {
-				break;
-			}
+			irqmp_add_eirq(cregs);
 		}
+	} else {
+		irqmp_add_eirq(regs);
 	}
 	return 0;
 }
@@ -230,6 +243,28 @@ int spi_fixup_prop(ambapp_core *core, struct prop_std *buf, int index)
 	}
 }
 
+
+#define ENTRY_OR_DEFAULT(array, idx, dflt) \
+	(idx < sizeof(array) / sizeof(array[0]) ? array[idx] : dflt)
+
+static int grgpio_irqgen_array[] = {
+	CONFIG_GRGPIO_IRQGEN
+};
+#define GRGPIO_NBITS(idx) \
+	ENTRY_OR_DEFAULT(grgpio_nbits_array, idx, 0);
+
+static int grgpio_nbits_array[] = {
+	CONFIG_GRGPIO_NBITS
+};
+#define GRGPIO_IMASKGEN(idx) \
+	ENTRY_OR_DEFAULT(grgpio_imaskgen_array, idx, 0x00000000);
+
+static int grgpio_imaskgen_array[] = {
+	CONFIG_GRGPIO_IMASKGEN
+};
+#define GRGPIO_IRQGEN(idx) \
+	ENTRY_OR_DEFAULT(grgpio_irqgen_array, idx, -1);
+
 #define GRGPIO_IMAP_MASK	0x1f1f1f1f
 
 #define GRGPIO_IRQMODE_STRAIGHT	1
@@ -239,9 +274,22 @@ int spi_fixup_prop(ambapp_core *core, struct prop_std *buf, int index)
 #define GRGPIO_MAX_PINS		32
 #define GRGPIO_IRQMAP_NOIRQ	0xffffffff
 
+struct grgpio_regs {
+	unsigned int data;		/* 0x00 */
+	unsigned int output;		/* 0x04 */
+	unsigned int dir;		/* 0x08 */
+	unsigned int imask;		/* 0x0c */
+	unsigned int ipolarity;		/* 0x10  */
+	unsigned int iedge;		/* 0x14 */
+	unsigned int bypass;		/* 0x18 */
+	unsigned int capability;	/* 0x1c */
+	unsigned int imap[8];		/* 0x20-0x3c */
+};
+
 struct grgpio_fixup_priv {
-	unsigned int base;
 	unsigned int nbits;
+	unsigned int imaskgen;
+	unsigned int irqmode;
 	unsigned int irqmap[GRGPIO_MAX_PINS];
 };
 
@@ -273,151 +321,134 @@ int grgpio_fixup_prop(ambapp_core *core, struct prop_std *buf, int index)
 	}
 }
 
-/* Probe grgpio for nbits and interrupts */
-int grgpio_fixup_core(ambapp_core *core)
+static int grgpio_irqgen_to_irqmode(int irqgen)
 {
-	int i, j;
-	int irqmode;
-	struct grgpio_regs {
-		unsigned int data;		/* 0x00 */
-		unsigned int output;		/* 0x04 */
-		unsigned int dir;		/* 0x08 */
-		unsigned int imask;		/* 0x0c */
-		unsigned int ipolarity;		/* 0x10  */
-		unsigned int iedge;		/* 0x14 */
-		unsigned int bypass;		/* 0x18 */
-		unsigned int __reserved;	/* 0x1c */
-		unsigned int imap[8];		/* 0x20-0x3c */
-	} *regs;
+	if (irqgen < 0)
+		return 0;
 
+	switch (irqgen) {
+	case 0:
+		return GRGPIO_IRQMODE_STRAIGHT;
+	case 1:
+		return GRGPIO_IRQMODE_SINGLE;
+	default:
+		return GRGPIO_IRQMODE_MAP;
+	}
+}
+
+/* Probes:
+ * - nbits, if not set
+ * - imaskgen, if not set
+ * - whether irqmode is IRQMODE_MAP for grgpiov1
+ */
+static void grgpio_probe(int idx, struct grgpio_regs *regs, struct grgpio_fixup_priv *gp)
+{
 	unsigned int imask;
-	unsigned int irqenabled;
+	unsigned int iedge;
 	unsigned int direction;
 	unsigned int output;
 	unsigned int lines;
+	unsigned int i;
 
-	/* Get registers */
-	struct grgpio_fixup_priv *gp = startup_malloc(sizeof(*gp));
-	ambapp_apbdev *dev = core->apbslv;
-
-	/* For keeping track of the running total of number of gpios over all
-	 * cores
-	 */
-	static unsigned int totalnbits = 0;
-
-	regs = (struct grgpio_regs *)dev->address;
-
-
-	/* Probe nbits, i.e. the number of lines */
-	direction = BYPASS_LOAD_PA(&regs->dir);
-	output = BYPASS_LOAD_PA(&regs->output);
-	BYPASS_STORE_PA(&regs->output, ~direction);
-	lines = direction | BYPASS_LOAD_PA(&regs->output);
-	BYPASS_STORE_PA(&regs->output, output);
-
-	/* Calculate nbits. Count the number of set bits in lines by
-	 * interativerly clearing the least significant bit */
-	for (gp->nbits = 0; lines; gp->nbits++)
-	{
-		lines &= lines - 1;
-	}
-	gp->base = totalnbits;
-	totalnbits += gp->nbits;
-
-	/* Probe which pins have irq enabled */
+	/* Prevent interrupts during probe */
 	imask = BYPASS_LOAD_PA(&regs->imask);
-	BYPASS_STORE_PA(&regs->imask, 0xffffffff);
-	irqenabled = BYPASS_LOAD_PA(&regs->imask);
+	BYPASS_STORE_PA(&regs->imask, 0x00000000);
+
+	if (!gp->nbits) {
+		/* Probe nbits, i.e. the number of lines */
+		direction = BYPASS_LOAD_PA(&regs->dir);
+		output = BYPASS_LOAD_PA(&regs->output);
+		BYPASS_STORE_PA(&regs->output, ~direction);
+		lines = direction | BYPASS_LOAD_PA(&regs->output);
+		BYPASS_STORE_PA(&regs->output, output);
+		for (i = 0; (lines & (1 << i)) && i < 32; i++)
+			;
+		gp->nbits = i;
+	}
+
+	if (!gp->imaskgen) {
+		/* Probe which pins have irq enabled */
+		iedge = BYPASS_LOAD_PA(&regs->iedge);
+		BYPASS_STORE_PA(&regs->imask, 0x00000000);
+		BYPASS_STORE_PA(&regs->iedge, 0xffffffff);
+		gp->imaskgen = BYPASS_LOAD_PA(&regs->iedge);
+		BYPASS_STORE_PA(&regs->iedge, iedge);
+	}
+
+	if (!gp->irqmode) {
+		unsigned int read;
+		unsigned int readback;
+
+		/* Probe if there are imap registers */
+		read = BYPASS_LOAD_PA(&regs->imap[0]);
+		BYPASS_STORE_PA(&regs->imap[0], ~read & GRGPIO_IMAP_MASK);
+		readback = BYPASS_LOAD_PA(&regs->imap[0]);
+		if(read != readback) {
+			gp->irqmode = GRGPIO_IRQMODE_MAP;
+			BYPASS_STORE_PA(&regs->imap[0], read);
+		}
+
+		/*
+		 * irqgen == 0 and irqgen == 1 cannot be distinguished for
+		 * version 1 of the core without going outside the grgpio
+		 * core. Therefore, this has to be configured from the outside
+		 * instead
+		 */
+		DBG_PRINTF1("grgpio%d: interrupts disabled - cannot distinguish irqgen 0 from 1\n", idx);
+	}
+
+	/* Restore original imask */
 	BYPASS_STORE_PA(&regs->imask, imask);
+}
 
-	irqmode = 0;
-#ifdef CONFIG_GRGPIO_SINGLEIRQ
-	/* irqgen == 1 cannot be probed without going outside the grgpio
-	 * core. This has to be configured from the outside instead */
-	irqmode = GRGPIO_IRQMODE_SINGLE;
-#else
+/* Sets up the irqmap as reported to Linux. Does not set up any registers */
+static void grgpio_setup_interrupts(ambapp_apbdev *dev,
+				    struct grgpio_regs *regs,
+				    struct grgpio_fixup_priv *gp)
+{
+	int pirq;
+	int i;
+	int j;
 
 
-	/* Probe if imap register address wrap around to general registers */
-	if (irqenabled) {
-		unsigned int oldval;
-
-		imask = BYPASS_LOAD_PA(&regs->imask);
-		oldval = BYPASS_LOAD_PA(&regs->imap[3]);
-		BYPASS_STORE_PA(&regs->imap[3], ~imask);
-		if (BYPASS_LOAD_PA(&regs->imask) != imask) {
-			irqmode = GRGPIO_IRQMODE_STRAIGHT;
-			BYPASS_STORE_PA(&regs->imask, imask);
-		}
-		BYPASS_STORE_PA(&regs->imap[3], oldval);
-	}
-
-	/* Probe if we have writable imap registers */
-	for (j = 0; j < 8 && !irqmode; j++) {
-		if ((0xf << (j*4)) & irqenabled) {
-			unsigned int read, readback;
-			void *address = &regs->imap[j];
-
-			read = BYPASS_LOAD_PA(address);
-			BYPASS_STORE_PA(address, ~read & GRGPIO_IMAP_MASK);
-			readback = BYPASS_LOAD_PA(address);
-			if(read != readback) {
-				irqmode = GRGPIO_IRQMODE_MAP;
-				BYPASS_STORE_PA(address, read);
-				break;
-			}
-		}
-	}
-	if(!irqmode) {
-		irqmode = GRGPIO_IRQMODE_STRAIGHT;
-	}
-#endif
-
-	/* Debug printout */
-	char *irqmodestr;
-	switch (irqmode) {
-	case GRGPIO_IRQMODE_STRAIGHT:
-		irqmodestr = "straight";
-		break;
-	case GRGPIO_IRQMODE_SINGLE:
-		irqmodestr = "single";
-		break;
-	case GRGPIO_IRQMODE_MAP:
-		irqmodestr = "map";
-		break;
-	default:
-		irqmodestr = "ERROR: undefined";
-		break;
-	}
+	int off = 0;
 
 	/* Fill in irq arrays */
-	int pirq = dev->irq[0];
-	int irqindex = 0;
+	pirq = dev->irq[0];
+	_memset(dev->irq, 0, sizeof(dev->irq));
 	for (i = 0; i < gp->nbits; i++) {
-		if (irqenabled & (1 << i)) {
-			if (irqmode == GRGPIO_IRQMODE_SINGLE) {
-				/* All irq enabled lines have index 0 */
-				gp->irqmap[i] = 0;
+		int irq = 0;
+
+		if (gp->irqmode && gp->imaskgen & (1 << i)) {
+			if (gp->irqmode == GRGPIO_IRQMODE_SINGLE) {
+				irq = pirq;
 			} else {
-				int off;
-				if (irqmode == GRGPIO_IRQMODE_MAP) {
-					/* All have index 0 */
+				/* int off; */
+				if (gp->irqmode == GRGPIO_IRQMODE_MAP) {
 					unsigned int map;
-					int j = i / 4;
+					j = i / 4;
 					map = BYPASS_LOAD_PA(&regs->imap[j]);
 					off = (map >> (8*(3-i))) & 0x1f;
 				} else {
 					off = i;
 				}
-				dev->irq[irqindex] = pirq + off;
-				gp->irqmap[i] = irqindex;
-				irqindex++;
+				irq = pirq + off;
 			}
+		}
+		if (irq) {
+			for (j = 0; dev->irq[j] && dev->irq[j] != irq; j++)
+				; /* Skip taken slots */
+			dev->irq[j] = irq;
+			gp->irqmap[i] = j;
 		} else {
 			gp->irqmap[i] = GRGPIO_IRQMAP_NOIRQ;
 		}
 	}
 
+	/*
+	 * Make sure that no irq are allowed to overlap with an eirq. This would
+	 * prevent Linux from even booting even if it is never triggered.
+	 */
 	for (i = 0; i < 32 && dev->irq[i] != 0; i++) {
 		/* Check if dev->irq[i] overlaps with an eirq */
 		for (j = 0; eirqs[j]; j++)
@@ -441,6 +472,60 @@ int grgpio_fixup_core(ambapp_core *core)
 		dev->irq[j-1] = 0;
 	}
 
+}
+
+/* Probe grgpio for nbits and interrupts */
+int grgpio_fixup_core(ambapp_core *core)
+{
+	struct grgpio_regs *regs;
+	struct grgpio_fixup_priv *gp;
+	ambapp_apbdev *dev;
+	int irqgen;
+
+	static unsigned int totalcores = 0;
+	int idx = totalcores++; /* core->subindex not initialized this early */
+
+	dev = core->apbslv;
+	regs = (struct grgpio_regs *)dev->address;
+	gp = startup_malloc(sizeof(*gp));
+
+	/* TODO: Make it possible for many cores to have different values */
+	gp->nbits = GRGPIO_NBITS(idx);
+	gp->imaskgen = GRGPIO_IMASKGEN(idx);
+	irqgen = GRGPIO_IRQGEN(idx);
+
+	if (core->apbslv->ver == 0) {
+		irqgen = 0;
+	} else if (core->apbslv->ver >= 2) {
+		unsigned int cap = BYPASS_LOAD_PA(&regs->capability);
+
+		gp->nbits = (cap & 0x1f) + 1;
+		irqgen = (cap >> 8) & 0x1f;
+	}
+	gp->irqmode = grgpio_irqgen_to_irqmode(irqgen);
+
+	if (CONFIG_GRGPIO_PROBE && (!gp->nbits || !gp->imaskgen || gp->irqmode))
+		grgpio_probe(idx, regs, gp);
+	grgpio_setup_interrupts(dev, regs, gp);
+
+#ifdef CONFIG_DEBUG
+	{
+		int i;
+		DBG_PRINTF3("grgpio%d:\n   nbits=0x%x, imaskgen=0x%x\n",
+			    idx, gp->nbits, gp->imaskgen);
+		DBG_PRINTF0("   irqs: ");
+		for (i = 0; dev->irq[i]; i++)
+			DBG_PRINTF2("%s0x%x", (i == 0 ? "" : ", "), dev->irq[i]);
+		DBG_PRINTF0("\n   interrupt map: ");
+		for (i = 0; i < gp->nbits; i++) {
+			if (gp->irqmap[i] < 32)
+				DBG_PRINTF2("%s0x%x", (i == 0 ? "" : ", "), gp->irqmap[i]);
+			else
+				DBG_PRINTF2("%s%s", (i == 0 ? "" : ", "), "-");
+		}
+		DBG_PRINTF0("\n");
+	}
+#endif
 	core->fixup_priv = gp;
 
 	return 0;
@@ -839,7 +924,6 @@ void ahb2ahb_bus_create_iommu(struct ambapp_bus *ainfo, struct ahb_bus *parent,
 			ambapp_ahbdev *bridge)
 {
 	struct ahb_bus *subbus;
-	struct node *n;
 	struct iommu_regs *regs = 0;
 	ambapp_ahbdev *pdev;
 	int j;
@@ -859,7 +943,7 @@ void ahb2ahb_bus_create_iommu(struct ambapp_bus *ainfo, struct ahb_bus *parent,
 	subbus->isiommu = 1;
 	
 	/* get per bus node */
-	n = ahb2ahb_create_bus_node(ainfo, 0, subbus->bus_idx);
+	ahb2ahb_create_bus_node(ainfo, 0, subbus->bus_idx);
 
 	/* scan for reg address */
 	for (j = 0; j < 4; j++) {
@@ -904,7 +988,6 @@ void ahb2ahb_bus_create(struct ambapp_bus *ainfo, struct ahb_bus *parent,
 	struct ahb_bus *subbus;
 	unsigned long ioarea;
 	int i;
-	struct node *n;
 	
 	/* Is there a PnP area behind the bridge?*/
 	ioarea = bridge->userdef[1];
@@ -934,7 +1017,7 @@ void ahb2ahb_bus_create(struct ambapp_bus *ainfo, struct ahb_bus *parent,
 	subbus->dir = (bridge->userdef[0] & AMBAPP_FLAG_FFACT_DIR);
 
 	/* get per bus node */
-	n = ahb2ahb_create_bus_node(ainfo, ioarea, subbus->bus_idx);
+	ahb2ahb_create_bus_node(ainfo, ioarea, subbus->bus_idx);
 	
 }
 
